import fire

from ktoolbox.cli import KToolBoxCli
from ktoolbox.utils import logger_init


def main():
    logger_init(cli_use=True)
    fire.Fire(KToolBoxCli)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import List, Union, Tuple

import aiofiles
from loguru import logger
from pathvalidate import sanitize_filename

from ktoolbox.action import ActionRet, fetch_all_creator_posts, FetchInterruptError
from ktoolbox.api.model import Post
from ktoolbox.api.posts import get_creator_post
from ktoolbox.configuration import config, PostStructureConfiguration
from ktoolbox.enum import PostFileTypeEnum, DataStorageNameEnum
from ktoolbox.job import Job, CreatorIndices

__all__ = ["create_job_from_post", "filter_posts_with_indices", "create_job_from_creator"]


async def create_job_from_post(
        post: Post,
        post_path: Path,
        *,
        post_structure: Union[PostStructureConfiguration, bool] = None,
        dump_post_data: bool = True
) -> List[Job]:
    """
    Create a list of download job from a post data

    :param post: post data
    :param post_path: Path of the post directory
    :param post_structure: post path structure, ``False`` -> disable, \
     ``True`` & ``None`` -> ``config.job.post_structure``
    :param dump_post_data: Whether to dump post data (post.json) in post directory
    """
    post_path.mkdir(exist_ok=True)

    # Load ``PostStructureConfiguration``
    if post_structure in [True, None]:
        post_structure = config.job.post_structure
    if post_structure:
        attachments_path = post_path / post_structure.attachments  # attachments
        attachments_path.mkdir(exist_ok=True)
        content_path = post_path / post_structure.content_filepath  # content
        content_path.parent.mkdir(exist_ok=True)
    else:
        attachments_path = post_path
        content_path = None

    # Create jobs
    jobs: List[Job] = []
    for attachment in post.attachments:  # attachments
        if not attachment.path:
            continue
        jobs.append(
            Job(
                path=attachments_path,
                alt_filename=attachment.name,
                server_path=attachment.path,
                type=PostFileTypeEnum.Attachment
            )
        )
    if post.file.path:  # file
        jobs.append(
            Job(
                path=post_path,
                alt_filename=f"{post.id}_{post.file.name}",
                server_path=post.file.path,
                type=PostFileTypeEnum.File
            )
        )

    # Write content file
    if content_path and post.content:
        async with aiofiles.open(content_path, "w", encoding=config.downloader.encoding) as f:
            await f.write(post.content)
    if dump_post_data:
        async with aiofiles.open(str(post_path / DataStorageNameEnum.PostData.value), "w", encoding="utf-8") as f:
            await f.write(
                post.json(indent=config.json_dump_indent)
            )

    return jobs


def filter_posts_with_indices(posts: List[Post], indices: CreatorIndices) -> Tuple[List[Post], CreatorIndices]:
    """
    Compare and filter posts by ``CreatorIndices`` data

    Only keep posts that was edited after last download.

    :param posts: Posts to filter
    :param indices: ``CreatorIndices`` data to use
    :return: A updated ``List[Post]`` and updated **new** ``CreatorIndices`` instance
    """
    new_list = list(
        filter(
            lambda x: x.id not in indices.posts or x.edited > indices.posts[x.id].edited, posts
        )
    )
    new_indices = indices.copy(deep=True)
    for post in new_list:
        new_indices.posts[post.id] = post
    return new_list, new_indices


async def create_job_from_creator(
        service: str,
        creator_id: str,
        path: Path,
        *,
        update_from: CreatorIndices = None,
        all_pages: bool = False,
        o: int = None,
        save_creator_indices: bool = True,
        mix_posts: bool = None
) -> ActionRet[List[Job]]:
    """
    Create a list of download job from a creator

    :param service: The service where the post is located
    :param creator_id: The ID of the creator
    :param path: The path for posts to download
    :param update_from: ``CreatorIndices`` data for update posts from current creator directory, \
     ``save_creator_indices`` will be enabled if this provided
    :param all_pages: Fetch all pages of posts, ``o`` will be ignored if enabled
    :param o: Result offset, stepping of 50 is enforced
    :param save_creator_indices: Record ``CreatorIndices`` data for update posts from current creator directory
    :param mix_posts: Save all files from different posts at same path, \
     ``update_from``, ``save_creator_indices`` will be ignored if enabled
    """
    mix_posts = config.job.mix_posts if mix_posts is None else mix_posts

    # Get posts
    logger.info(f"Start fetching posts from creator {creator_id}")
    if all_pages:
        post_list: List[Post] = []
        try:
            async for part in fetch_all_creator_posts(service=service, creator_id=creator_id):
                post_list += part
        except FetchInterruptError as e:
            return ActionRet(**e.ret.dict())
    else:
        ret = await get_creator_post(service=service, creator_id=creator_id, o=o)
        if ret:
            post_list = ret.data
        else:
            return ActionRet(**ret.dict())
    logger.info(f"Get {len(post_list)} posts, start creating jobs")

    # Filter posts and generate ``CreatorIndices``
    if not mix_posts:
        indices = None
        if update_from:
            post_list, indices = filter_posts_with_indices(post_list, update_from)
            logger.info(f"{len(post_list)} posts will be downloaded")
        elif save_creator_indices:  # It's unnecessary to create indices again when ``update_from`` was provided
            indices = CreatorIndices(
                creator_id=creator_id,
                service=service,
                posts={post.id: post for post in post_list},
                posts_path={post.id: path / sanitize_filename(post.title) for post in post_list}
            )
        if indices:
            async with aiofiles.open(
                    path / DataStorageNameEnum.CreatorIndicesData.value,
                    "w",
                    encoding="utf-8"
            ) as f:
                await f.write(indices.json(indent=config.json_dump_indent))

    job_list: List[Job] = []
    for post in post_list:
        # Get post path
        if mix_posts:
            default_post_path = path
        elif config.job.post_id_as_path:
            default_post_path = path / post.id
        else:
            default_post_path = path / sanitize_filename(post.title)
        if update_from:
            if not (post_path := update_from.posts_path.get(post.id)):
                post_path = default_post_path
        else:
            post_path = default_post_path

        # Generate jobs
        job_list += await create_job_from_post(
            post=post,
            post_path=post_path,
            post_structure=False if mix_posts else None,
            dump_post_data=not mix_posts
        )
    return ActionRet(data=job_list)



import cgi
import logging
import sys
import urllib.parse
from pathlib import Path
from typing import Generic, TypeVar, Optional, Dict, List, Tuple

import aiofiles
from loguru import logger
from pydantic import BaseModel

from ktoolbox.configuration import config
from ktoolbox.enum import RetCodeEnum, DataStorageNameEnum
from ktoolbox.model import SearchResult

__all__ = ["BaseRet", "filename_from_headers", "generate_msg", "logger_init", "dump_search", "parse_webpage_url"]

_T = TypeVar('_T')


class BaseRet(BaseModel, Generic[_T]):
    """Base data model of function return value"""
    code: int = RetCodeEnum.Success.value
    message: str = ''
    exception: Optional[Exception] = None
    data: Optional[_T] = None

    class Config(BaseModel.Config):
        arbitrary_types_allowed = True

    def __bool__(self):
        return self.code == RetCodeEnum.Success


def parse_header(line: str) -> Dict[str, Optional[str]]:
    """
    Alternative resolution for parsing header line.

    Apply when ``cgi.parse_header`` is unable to use due to the deprecation of `cgi` module.

    https://peps.python.org/pep-0594/#cgi

    :param line: Header line
    :return: Dict of header line

    .. rubric:: Example:
    ```
    parse_header("text/html; charset=utf-8")
    ```

    .. rubric:: Return:
    ```
    {'text/html': None, 'charset': 'utf-8'}
    ```
    """
    dict_value: Dict[str, Optional[str]] = {}
    for item in line.split(";"):
        if len(pair := item.split("=")) == 1:
            dict_value[pair[0]] = None
        else:
            dict_value.setdefault(*pair)
    return dict_value


def filename_from_headers(headers: Dict[str, str]) -> Optional[str]:
    """
    Get file name from headers.

    Parse from ``Content-Disposition``.

    :param headers: HTTP headers
    :return: File name

    .. rubric:: Example:
    ```
    filename_from_headers('attachment;filename*=utf-8\\'\\'README%2Emd;filename="README.md"')
    ```

    .. rubric:: Return:
    ```
    README.md
    ```
    """
    if not (disposition := headers.get("Content-Disposition")):
        if not (disposition := headers.get("content-disposition")):
            return None
    _, options = cgi.parse_header(disposition)  # alternative: `parse_header` in `utils.py`
    if filename := options.get("filename*"):
        if len(name_with_charset := filename.split("''")) == 2:
            charset, name = name_with_charset
            return urllib.parse.unquote(name, charset)
    if filename := options.get("filename"):
        return urllib.parse.unquote(filename, config.downloader.encoding)
    return None


def generate_msg(title: str = None, **kwargs):
    """
    Generate message for ``BaseRet`` and logger

    :param title: Message title
    :param kwargs: Extra data
    """
    title: str = title or ""
    return f"{title} - {kwargs}" if kwargs else title


def logger_init(cli_use: bool = False, disable_stdout: bool = False):
    """
    Initialize ``loguru`` logger

    :param cli_use: Set logger level ``INFO`` and filter out ``SUCCESS``
    :param disable_stdout: Disable default output stream
    """
    if disable_stdout:
        logger.remove()
    elif cli_use:
        logger.remove()
        logger.add(
            sys.stderr,
            level=logging.INFO,
            filter=lambda record: record["level"].name != "SUCCESS"
        )
    if path := config.logger.path:
        path.mkdir(exist_ok=True)
        if path is not None:
            logger.add(
                path / DataStorageNameEnum.LogData.value,
                level=config.logger.level,
                rotation=config.logger.rotation,
                diagnose=True
            )


async def dump_search(result: List[BaseModel], path: Path):
    async with aiofiles.open(str(path), "w", encoding="utf-8") as f:
        await f.write(
            SearchResult(result=result)
            .json(indent=config.json_dump_indent)
        )


def parse_webpage_url(url: str) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    # noinspection SpellCheckingInspection
    """
    Fetch **service**, **user_id**, **post_id** from webpage url

    Each part can be ``None`` if not found in url.

    :param url: Kemono Webpage url
    :return: Tuple of **service**, **user_id**, **post_id**
    """
    path_url = Path(url)
    parts = path_url.parts
    if (url_parts_len := len(parts)) < 7:
        # Pad to full size
        parts += tuple(None for _ in range(7 - url_parts_len))
    _scheme, _netloc, service, _user_key, user_id, _post_key, post_id = parts
    return service, user_id, post_id



import datetime
import logging
from pathlib import Path
from typing import Literal, Union, Optional

from pydantic import BaseModel, BaseSettings

__all__ = [
    "config",
    "APIConfiguration",
    "DownloaderConfiguration",
    "PostStructureConfiguration",
    "JobConfiguration",
    "Configuration"
]


# noinspection SpellCheckingInspection
class APIConfiguration(BaseModel):
    """Kemono API Configuration"""
    scheme: Literal["http", "https"] = "https"
    """Kemono API URL scheme"""
    netloc: str = "kemono.su"
    """Kemono API URL netloc"""
    statics_netloc: str = "img.kemono.su"
    """URL netloc of Kemono server for static files (e.g. images)"""
    files_netloc: str = "kemono.su"
    """URL netloc of Kemono server for post files"""
    path: str = "/api/v1"
    """Kemono API URL root path"""
    timeout: float = 5.0
    """API request timeout"""
    retry_times: int = 3
    """API request retry times (when request failed)"""
    retry_interval: float = 2.0
    """Seconds of API request retry interval"""


class DownloaderConfiguration(BaseModel):
    """File Downloader Configuration"""
    scheme: Literal["http", "https"] = "https"
    """Downloader URL scheme"""
    timeout: float = 30.0
    """Downloader request timeout"""
    encoding: str = "utf-8"
    """Charset for filename parsing and post content text saving"""
    buffer_size: int = 20480
    """Number of bytes of file I/O buffer for each downloading file"""
    chunk_size: int = 1024
    """Number of bytes of chunk of downloader stream"""
    temp_suffix: str = "tmp"
    """Temp filename suffix of downloading files"""


class PostStructureConfiguration(BaseModel):
    # noinspection SpellCheckingInspection
    """
    Post path structure model

    .. rubric:: Default:
    |__+ ..

    |__+ attachments

    |____+ (e.g. 1.png)

    |____+ (e.g. 2.png)

    |__+ content.txt

    |__+ <Post file>

    |__+ <Post data (post.ktoolbox.json)>
    """
    attachments: Path = Path("attachments")
    """Sub path of attachment directory"""
    content_filepath: Path = Path("index.html")
    """Sub path of post content HTML file"""


class JobConfiguration(BaseModel):
    """Download jobs Configuration"""
    count: int = 4
    """Number of coroutines for concurrent download"""
    post_id_as_path: bool = False
    """Use post ID as post directory name"""
    post_structure: PostStructureConfiguration = PostStructureConfiguration()
    """Post path structure"""
    mix_posts: bool = False
    """
    Save all files from different posts at same path in creator directory

    .. rubric:: NOTICE:
    It would not create any post directory, and ``CreatorIndices`` would not been recorded, \
    without ``CreatorIndices`` you **cannot update** the creator directory.
    """
    job_list_filepath: Optional[Path] = None
    """Filepath for job list data saving, ``None`` for disable job list saving"""


class LoggerConfiguration(BaseModel):
    """Logger configuration"""
    path: Optional[Path] = Path("logs")
    """Path to save logs, ``None`` for disable log file output"""
    level: Union[str, int] = logging.DEBUG
    """Log filter level"""
    rotation: Union[str, int, datetime.time, datetime.timedelta] = "1 week"
    """Log rotation"""


class Configuration(BaseSettings):
    """KToolBox Configuration"""
    api: APIConfiguration = APIConfiguration()
    downloader: DownloaderConfiguration = DownloaderConfiguration()
    job: JobConfiguration = JobConfiguration()
    logger: LoggerConfiguration = LoggerConfiguration()

    ssl_verify: bool = True
    # noinspection SpellCheckingInspection
    """Enable SSL certificate verification for Kemono API server and download server"""
    json_dump_indent: int = 4
    """Indent of JSON file dump"""

    class Config(BaseSettings.Config):
        env_prefix = 'ktoolbox_'
        env_nested_delimiter = '__'
        env_file = '.env'
        env_file_encoding = 'utf-8'


config = Configuration(_env_file='prod.env')
